/**************************************************************************************************
* NOTICE OF COPYRIGHT AND OWNERSHIP OF SOFTWARE:
*
* Copyright 2010 Flexera Software, Inc.  All Rights Reserved.  
*
* This computer program is the property of Flexera Software, Inc.  Any use, copy, publication, 
* distribution, display, modification, or transmission of this computer program in whole or in part
* in any form or by any means without the prior express written permission of Flexera Software, 
* Inc. is strictly prohibited.
* 
* Except when expressly provided by Flexera Software, Inc. in writing, possession of this computer 
* program shall not be construed to confer any license or rights under any of Flexera Software,
* Inc.'s intellectual property rights, whether by estoppel, implication, or otherwise.  
* 
* ALL COPIES OF THIS PROGRAM MUST DISPLAY THIS NOTICE OF COPYRIGHT AND OWNERSHIP IN FULL.
**************************************************************************************************/
/*
	This source file is part of a sample utility that demonstrates Composite Transactions.

	It declares the functions that operate on fulfillment records (view, delete, repair).

	Access to fullfillments is supported by the common API and is not specific to composite 
	transactions; the fulfillments may also have been created by local activations, single action
	trnsactions or short code transactions.

	The common API uses two objects for representing data in trusted storage:

		FlxActProdLicSpc	One fulfillment record.
		FlxActLicSpc		A collection of FlxActProdLicSpc's 

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FlxActCommon.h"
#include "FlxActError.h"

#include "compTranUtilFulfillments.h"

typedef struct FlxCtuFulfillmentFilter_struct
{
	FlxActBool	bViewAllIds;
	FlxActBool	bViewAllProducts;
	FlxActBool	bViewAllFeatures;
	FlxActBool	bViewTrusted;
	FlxActBool	bViewUntrusted;
	FlxActBool	bViewEnabled;
	FlxActBool	bViewDisabled;
	const char*	pszFidFilter;
	const char*	pszPidFilter;
	const char*	pszFeatureFilter;

	uint32_t	totalCount;
	uint32_t	passCount;
} FlxCtuFulfillmentFilter;

typedef enum FlxCtuViewType_enum
{
	flxCtuViewTypeShort	= 0,
	flxCtuViewTypeLong,
	flxCtuViewTypeLongWithCounts,
} FlxCtuViewType;

typedef FlxActBool (*FilterCallback)(FlxCtuContext*		pContext,
									 FlxActLicSpc		licSpec, 
									 FlxActProdLicSpc	prodSpec,
									 void*				pData);

static FlxActBool sAddRepairAction(FlxCtuContext*	pContext,
								   FlxActLicSpc		licSpec, 
								   FlxActProdLicSpc	prodSpec,
								   void*			pData);

static FlxActBool sDeleteFulfillment(FlxCtuContext*		pContext,
									 FlxActLicSpc		licSpec, 
									 FlxActProdLicSpc	prodSpec,
									 void*				pData);

static FlxActBool sForEachFilterPass(FlxCtuContext*				pContext, 
									 FlxCtuFulfillmentFilter*	pFilter,
									 FilterCallback				callbackFn,
									 void*						pData);

static const char* sGetFulfillmentTypeString(FlxActProdLicSpc prodSpec);
static const char* sGetFulfillmentTrustFlagsString(FlxActProdLicSpc prodSpec);
static uint32_t sGetNonOverdraftCountTotal(FlxActProdLicSpc prodSpec);
static void		  sInitFilterCounts(FlxCtuFulfillmentFilter* pFilter);
static void       sInitFilterAll(FlxCtuFulfillmentFilter* pFilter);
static FlxActBool sIsFilterPass(FlxActProdLicSpc prodSpec, const FlxCtuFulfillmentFilter* pFilter);
static FlxActBool sHasThisFeatureToken(FlxActProdLicSpc prodSpec, const char* pszFeatureName);
static FlxActBool sHasThisFulfillmentId(FlxActProdLicSpc prodSpec, const char* pszFulfillmentId);
static FlxActBool sHasThisProductId(FlxActProdLicSpc prodSpec, const char* pszProductId);
static FlxCtuViewType sGetViewType(const FlxCtuContext* pContext, const FlxCtuCommand* pCommand);
static void sSetFilter(const FlxCtuCommand* pCommand, FlxCtuFulfillmentFilter* pFilter);
static void sViewFilteredFulfillments(FlxCtuContext*			pContext, 
									  FlxCtuFulfillmentFilter*	pFilter,
									  FlxCtuViewType			viewType);
static FlxActBool sViewFulfillment(FlxCtuContext*	pContext,
								   FlxActLicSpc		licSpec, 
								   FlxActProdLicSpc	prodSpec,
								   void*			pData);
static void sViewFulfillmentShort(FlxActProdLicSpc prodSpec);
static void sViewFulfillmentLong(FlxCtuContext* pContext, FlxActProdLicSpc prodSpec);
static void sViewFulfillmentCounts(FlxCtuContext* pContext, FlxActProdLicSpc prodSpec);
static void sViewVendorDictionary(FlxCtuContext* pContext, FlxActProdLicSpc prodSpec);
static void sViewNameAndFeatures(const char* pszName, FlxActProdLicSpc prodSpec);
static void sViewNameAndValue(const char* pszName, const char* pszValue);
static void sViewNameAndUint32(const char* pszName, uint32_t value);


typedef struct FrStateCallbackData_struct
{
	FlxActBool bIsSet;
	uint32_t   trustFlags;
	FlxActBool bIsDisabled;
} FrStateCallbackData;

static FlxActBool sFrStateCallback(FlxCtuContext*	pContext,
								   FlxActLicSpc		licSpec, 
								   FlxActProdLicSpc	prodSpec,
								   void*			pData);

/***************************************************************************************************
*	flxCtuDoCommandDeleteFR
*
* Process a deleteFR command - delete fulfillment record(s).
***************************************************************************************************/
FlxActBool flxCtuDoCommandDeleteFR(FlxCtuContext* pContext, const FlxCtuCommand* pCommand)
{
	FlxCtuFulfillmentFilter	filter;

	/* Set the filtering choices from the command value and attributes. */
	sSetFilter(pCommand, &filter);

	/* Give the user a preview of the FRs that will be deleted. */
	sViewFilteredFulfillments(pContext, &filter, flxCtuViewTypeShort);

	if (filter.totalCount == 0)
		flxCtuPrintMessage("\nNo fulfillment records in trusted storage.\n");
	else if (filter.passCount == 0)
		flxCtuPrintMessage("No records to delete.\n");
	else
	{
		if (flxCtuContextHasCommand(pContext, CMD_BRIEF) ||
			flxCtuConfirm("\n%u record%s to delete. Confirm?\n", 
						  filter.passCount, 
						  (filter.passCount == 1)? "" : "s"))
		{
			/* Call sDeleteFulfillment for each FR that meets the filter criteria. */
			if ( !sForEachFilterPass(pContext, &filter, sDeleteFulfillment, NULL))
				return FLX_ACT_FALSE;
		}
	}
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	flxCtuDoCommandView
*
* Process a view command - view fulfillment records.
***************************************************************************************************/
FlxActBool flxCtuDoCommandView(FlxCtuContext* pContext, const FlxCtuCommand* pCommand)
{
	FlxCtuViewType			viewType;
	FlxCtuFulfillmentFilter	filter;

	flxCtuPrintMessage("Viewing %s trusted storage.\n",
					   (pContext->bIsServer)? "server" : "application");

	/* Get the filtering choices from the command value and attributes. */
	sSetFilter(pCommand, &filter);

	/* Get the view type (short, long counts) from the command attributes. */
	viewType = sGetViewType(pContext, pCommand);

	/* View the fulfillments that meet the filter criteria. */
	sViewFilteredFulfillments(pContext, &filter, viewType);

	if (filter.totalCount == 0)
		flxCtuPrintMessage("\nNo fulfillment records in trusted storage.\n");
	else
		flxCtuPrintMessage("\nViewed %d of %d fulfillment records.\n", 
						   filter.passCount,
						   filter.totalCount);

	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	flxCtuFulfillmentsAddRepairActions
*
* Add a repair action to the current response for each fulfillment record that is not fully trusted.
***************************************************************************************************/
FlxActBool flxCtuFulfillmentsAddRepairActions(FlxCtuContext* pContext, uint32_t* pCount)
{
	FlxActTranRequest		hTranRequest;
	FlxCtuFulfillmentFilter filter;

	/* Get the current request. */
	if ( !flxActTransactionGetRequest(pContext->hTransaction, &hTranRequest) )
		FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	/* Create a filter for untrusted FRs (exclude trusted ones). */
	sInitFilterAll(&filter);
	filter.bViewTrusted = FLX_ACT_FALSE;

	/* Call sAddRepairAction for each untrusted FR. */
	if ( !sForEachFilterPass(pContext, &filter, sAddRepairAction, NULL))
		return FLX_ACT_FALSE;

	/* Return the pass count to the caller. */
	*pCount = filter.passCount;
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	flxCtuFulfillmentGetState
*
* Return the trusted and enabled state of an FR.
***************************************************************************************************/
FlxActBool flxCtuFulfillmentGetState(FlxCtuContext* pContext, 
									 const char*	pszFulfillmentId,
									 FlxActBool*	pbIsTrusted, 
									 FlxActBool*	pbIsEnabled)
{
	FlxCtuFulfillmentFilter filter;
	FrStateCallbackData     stateData;

	/* Set up a filter for the required FulfillmentId. */
	sInitFilterAll(&filter);
	filter.bViewAllIds  = FLX_ACT_FALSE;
	filter.pszFidFilter = pszFulfillmentId;

	/* Initialise the state data in case the Fr is not found. */
	stateData.bIsSet = FLX_ACT_FALSE;

	/* Set the state for that Fulfillment. */
	sForEachFilterPass(pContext, &filter, sFrStateCallback, (void*)&stateData);

	/* Check the Fr was found. */
	if ( !stateData.bIsSet)
		return FLX_ACT_FALSE;

	/* Return the state. */
	*pbIsEnabled = (stateData.bIsDisabled)? FLX_ACT_FALSE : FLX_ACT_TRUE;
	*pbIsTrusted = (stateData.trustFlags == FLAGS_FULLY_TRUSTED)? FLX_ACT_TRUE : FLX_ACT_FALSE;
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sAddRepairAction
*
* Add a repair action for this fulfillment to the current request.
***************************************************************************************************/
static FlxActBool sAddRepairAction(FlxCtuContext*	pContext,
								   FlxActLicSpc		licSpec, 
								   FlxActProdLicSpc	prodSpec,
								   void*			pData)
{
	FlxActTranRequest	hTranRequest;
	const char*			pszFulfillmentId;
	FlxActTranReqAction	hAction;

	/* Get the current request. */
	if ( !flxActTransactionGetRequest(pContext->hTransaction, &hTranRequest) )
		FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	/* Get the fulfillment id. */
	pszFulfillmentId = flxActCommonProdLicSpcFulfillmentIdGet(prodSpec);

	/* Create a new repair action and add it to the request. */
	if ( !flxActTranRequestAddAction(hTranRequest, FLX_ACT_TRAN_REQ_ACTION_REPAIR, &hAction) )
		FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	/* Specify this FR by setting the action's FulfillmentId attribute. */
	if ( !flxActTranReqActionSetAttribute(hAction, 
										  FLX_ACT_TRAN_REQ_ACT_FULFILLMENT_ID, 
										  pszFulfillmentId) )
	{
		FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);
	}

	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sDeleteFR
*
* Delete an FR (by deleting the prodSpec from its licSpec container).
***************************************************************************************************/
static FlxActBool sDeleteFulfillment(FlxCtuContext*		pContext,
									 FlxActLicSpc		licSpec, 
									 FlxActProdLicSpc	prodSpec,
									 void*				pData)
{
	flxCtuPrintMessage("Deleting fulfillment %s\n", 
					   flxActCommonProdLicSpcFulfillmentIdGet(prodSpec));

	if ( !flxActCommonLicSpcProductDelete(licSpec, prodSpec))
	{
		flxCtuContextPrintLastApiFlexError(pContext);
		return FLX_ACT_FALSE;
	}
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sForEachFilterPass
*
* Call callbackFn for each FR that passes the filter criteria.
***************************************************************************************************/
static FlxActBool sForEachFilterPass(FlxCtuContext*				pContext, 
									 FlxCtuFulfillmentFilter*	pFilter,
									 FilterCallback				callbackFn,
									 void*						pData)
{
	FlxActLicSpc		licSpec		= NULL;		/* A container for FlxActProdLicSpc's. */
	FlxActProdLicSpc	prodSpec	= NULL;		/* Represents one FR. */
	uint32_t			index;

	/* Create the container. */
	if ( !flxActCommonLicSpcCreate(pContext->hApi, &licSpec) )
	{
		flxCtuContextPrintLastApiFlexError(pContext);
		return FLX_ACT_FALSE;
	}
	
	/* Fill the container with all FRs. */
	if ( !flxActCommonLicSpcPopulateAllFromTS(licSpec) )
	{
		flxCtuContextPrintLastApiFlexError(pContext);
		flxActCommonLicSpcDelete(licSpec);
		return FLX_ACT_FALSE;
	}

	/* Initialise the counts. */
	pFilter->totalCount = flxActCommonLicSpcGetNumberProducts(licSpec);
	pFilter->passCount  = 0;

	/* Iterate over the fulfillments. */
	for (index = 0; index < pFilter->totalCount; index++)
	{
		prodSpec = flxActCommonLicSpcGet(licSpec, index);
		if (sIsFilterPass(prodSpec, pFilter))
		{	/* Meets filter criteria - call callbackFn.	*/
			if ( !(callbackFn)(pContext, licSpec, prodSpec, pData) )
			{
				flxActCommonLicSpcDelete(licSpec);
				return FLX_ACT_FALSE;
			}
			pFilter->passCount++;
		}
	}
	flxActCommonLicSpcDelete(licSpec);
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sFrStateCallback
*
* Sets the FR state in the pData (FrStateCallbackData).
***************************************************************************************************/
static FlxActBool sFrStateCallback(FlxCtuContext*	pContext,
								   FlxActLicSpc		licSpec, 
								   FlxActProdLicSpc	prodSpec,
								   void*			pData)
{
	FrStateCallbackData* pStateData = (FrStateCallbackData*)pData;

	pStateData->trustFlags  = flxActCommonProdLicSpcTrustFlagsGet(prodSpec);
	pStateData->bIsDisabled = flxActCommonProdLicSpcIsDisabled(prodSpec);
	pStateData->bIsSet	  = FLX_ACT_TRUE;
	return FLX_ACT_FALSE;	/* No need to continue the iteration. */
}

/***************************************************************************************************
*	sGetFulfillmentTrustFlagsString
*
* Return a printable string for the trust flags of a fulfillment record.
***************************************************************************************************/
static const char* sGetFulfillmentTrustFlagsString(FlxActProdLicSpc prodSpec)
{
	uint32_t trustFlags = flxActCommonProdLicSpcTrustFlagsGet(prodSpec);
	static char szTrustString[32];

	if (trustFlags == FLAGS_FULLY_TRUSTED)
	{
		sprintf(szTrustString, "0x%X Fully Trusted", trustFlags);
	}
	else
	{
		sprintf(szTrustString, 
			"0x%X **BROKEN** %s%s%s", 
			trustFlags,
			((trustFlags & FLAGS_TRUST_RESTORE) == 0)? "Restore " : "",
			((trustFlags & FLAGS_TRUST_HOST)    == 0)? "Host "    : "",
			((trustFlags & FLAGS_TRUST_TIME)    == 0)? "Time "    : ""
			);
	}
	return szTrustString;
}

/***************************************************************************************************
*	sGetFulfillmentTypeString
*
* Return a printable string for the type of a fulfillment record.
***************************************************************************************************/
static const char* sGetFulfillmentTypeString(FlxActProdLicSpc prodSpec)
{
	static const char *szType[] = {	"0 Unknown", 
									"1 Trial", 
									"2 Shortcode",  
									"3 Emergency", 
									"4 Served Activation", 
									"5 Served Overdraft Activation",
									"6 Publisher Activation", 
									"7 Publisher Overdraft Activation"};

	FlxFulfillmentType fulfillmentType = flxActCommonProdLicSpcFulfillmentTypeGet(prodSpec);
	if ( fulfillmentType >= (sizeof(szType) / sizeof(szType[0])) )
		fulfillmentType = FULFILLMENT_TYPE_UNKNOWN;
	return szType[fulfillmentType];
}

/***************************************************************************************************
*	sGetNonOverdraftCountTotal
*
* Return the sum of the basic counts.
***************************************************************************************************/
static uint32_t sGetNonOverdraftCountTotal(FlxActProdLicSpc prodSpec)
{
	return	flxActCommonProdLicSpcCountGet(prodSpec, CONCURRENT)
		  +	flxActCommonProdLicSpcCountGet(prodSpec, ACTIVATABLE)
		  +	flxActCommonProdLicSpcCountGet(prodSpec, HYBRID);
}

/***************************************************************************************************
*	sGetViewType
*
* Returns the format of the view output from the command attributes.
***************************************************************************************************/
static FlxCtuViewType sGetViewType(const FlxCtuContext* pContext, const FlxCtuCommand* pCommand)
{
	FlxCtuViewType viewType = flxCtuViewTypeShort;
	if (flxCtuCommandHasAttribute(pCommand, CMD_ATTR_FORMAT))
	{
		if (flxCtuCommandAttributeIs(pCommand, CMD_ATTR_FORMAT, ATTR_VALUE_LONG))
		{
			viewType = flxCtuViewTypeLong;

			/* Shows counts if viewing server TS or if command attribute "counts=yes" */
			if (flxCtuCommandAttributeIsYes(pCommand, "counts", pContext->bIsServer))
				viewType = flxCtuViewTypeLongWithCounts;
		}
	}
	return viewType;
}

/***************************************************************************************************
*	sInitFilterAll
*
* Initialise the filter so that all records will pass.
***************************************************************************************************/
static void sInitFilterAll(FlxCtuFulfillmentFilter* pFilter)
{
	pFilter->bViewAllIds		= FLX_ACT_TRUE;
	pFilter->bViewAllProducts	= FLX_ACT_TRUE;
	pFilter->bViewAllFeatures	= FLX_ACT_TRUE;
	pFilter->bViewTrusted		= FLX_ACT_TRUE;
	pFilter->bViewUntrusted		= FLX_ACT_TRUE;
	pFilter->bViewEnabled		= FLX_ACT_TRUE;
	pFilter->bViewDisabled		= FLX_ACT_TRUE;

	pFilter->pszFidFilter		= "";
	pFilter->pszPidFilter		= "";
	pFilter->pszFeatureFilter	= "";

	sInitFilterCounts(pFilter);
}

/***************************************************************************************************
*	sInitFilterCounts
*
* Initialise the pass and total counts.
***************************************************************************************************/
static void sInitFilterCounts(FlxCtuFulfillmentFilter* pFilter)
{
	pFilter->totalCount = 0;
	pFilter->passCount  = 0;
}

/***************************************************************************************************
*	sIsFilterPass
*
* Returns FLX_ACT_TRUE if the record passes the filter criteria.
***************************************************************************************************/
static FlxActBool sIsFilterPass(FlxActProdLicSpc prodSpec, const FlxCtuFulfillmentFilter* pFilter)
{
	FlxActBool bIsPass = FLX_ACT_FALSE;

	if (pFilter->bViewAllIds || sHasThisFulfillmentId(prodSpec, pFilter->pszFidFilter))
	{
		/* Passes FulfillmentId filter. */
		if (pFilter->bViewAllProducts || sHasThisProductId(prodSpec, pFilter->pszPidFilter))
		{
			/* Passes ProductId filter. */
			if (   pFilter->bViewAllFeatures 
				|| sHasThisFeatureToken(prodSpec, pFilter->pszFeatureFilter))
			{
				/* Passes Feature filter. */
				uint32_t trustFlags = flxActCommonProdLicSpcTrustFlagsGet(prodSpec);
				if (   (pFilter->bViewTrusted   && (trustFlags == FLAGS_FULLY_TRUSTED))
					|| (pFilter->bViewUntrusted && (trustFlags != FLAGS_FULLY_TRUSTED)) )
				{
					/* Passes Trusted filter. */
					FlxActBool bIsDisabled = flxActCommonProdLicSpcIsDisabled(prodSpec);
					if (   (pFilter->bViewEnabled  && !bIsDisabled)
						|| (pFilter->bViewDisabled &&  bIsDisabled) )
					{
						/* Passes Disabled filter. */
						bIsPass = FLX_ACT_TRUE;
					}
				}
			}
		}
	}
	return bIsPass;
}

/***************************************************************************************************
*	sHasThisFeatureToken 
*
* Returns FLX_ACT_TRUE if the token appears anywhere in the feature line. 
***************************************************************************************************/
static FlxActBool sHasThisFeatureToken(FlxActProdLicSpc prodSpec, const char* pszFeatureToken)
{
	const char* pszFeatureLine = flxActCommonProdLicSpcFeatureLineGet(prodSpec);

	if (pszFeatureLine == NULL)
		return FLX_ACT_FALSE;

	return (strstr(pszFeatureLine, pszFeatureToken) == NULL)? FLX_ACT_FALSE : FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sHasThisFulfillmentId
*
* Returns FLX_ACT_TRUE if the record has the supplied FulfillmentId. 
***************************************************************************************************/
static FlxActBool sHasThisFulfillmentId(FlxActProdLicSpc prodSpec, const char* pszFulfillmentId)
{
	return (0 == strcmp(pszFulfillmentId, flxActCommonProdLicSpcFulfillmentIdGet(prodSpec)));
}

/***************************************************************************************************
*	sHasThisProductId
*
* Returns FLX_ACT_TRUE if the record has the supplied ProductId. 
***************************************************************************************************/
static FlxActBool sHasThisProductId(FlxActProdLicSpc prodSpec, const char* pszProductId)
{
	const char* pszFrProductId = flxActCommonProdLicSpcProductIdGet(prodSpec);

	if (pszFrProductId == NULL)
		return FLX_ACT_FALSE;

	return (0 == strcmp(pszProductId, pszFrProductId));
}

/***************************************************************************************************
*	sSetFilter
*
* Set the filter criteria from the command value and attributes. 
***************************************************************************************************/
static void sSetFilter(const FlxCtuCommand* pCommand, FlxCtuFulfillmentFilter* pFilter)
{
	pFilter->bViewAllIds		= !flxCtuCommandHasValue(pCommand);
	pFilter->bViewAllProducts	= !flxCtuCommandHasAttribute(pCommand, CMD_ATTR_PID);
	pFilter->bViewAllFeatures	= !flxCtuCommandHasAttribute(pCommand, CMD_ATTR_FEATURE);
	pFilter->bViewTrusted		= !flxCtuCommandAttributeIs(pCommand, CMD_ATTR_TRUSTED, "no");
	pFilter->bViewUntrusted		= !flxCtuCommandAttributeIs(pCommand, CMD_ATTR_TRUSTED, "yes");
	pFilter->bViewEnabled		= !flxCtuCommandAttributeIs(pCommand, CMD_ATTR_ENABLED, "no");
	pFilter->bViewDisabled		= !flxCtuCommandAttributeIs(pCommand, CMD_ATTR_ENABLED, "yes");

	pFilter->pszFidFilter		= pFilter->bViewAllIds? "" : flxCtuCommandGetValue(pCommand);
	pFilter->pszPidFilter		= pFilter->bViewAllProducts?
								  "" : flxCtuCommandGetAttribute(pCommand, CMD_ATTR_PID);
	pFilter->pszFeatureFilter	= pFilter->bViewAllFeatures?
								  "" : flxCtuCommandGetAttribute(pCommand, CMD_ATTR_FEATURE);
	sInitFilterCounts(pFilter);
}

/***************************************************************************************************
*	sViewFilteredFulfillments
*
* Print details of the fulfillments that pass the filter criteria.
***************************************************************************************************/
static void sViewFilteredFulfillments(FlxCtuContext*			pContext, 
									  FlxCtuFulfillmentFilter*	pFilter,
									  FlxCtuViewType			viewType)
{
	/* If short view, print column headings. */
	if (viewType == flxCtuViewTypeShort)
	{
		flxCtuPrintMessage(
			"\n"
			"U if untrusted\n"
			". D if disabled\n"
			". . fulfillmentType\n" 
			". . . Expiry   Count FulfillmentId                   ProductId\n"
			"\n");
	}

	/* Call sViewFulfillment for each FR that meets the filter criteria. */
	(void)sForEachFilterPass(pContext, pFilter, sViewFulfillment, (void*)&viewType);
}

/***************************************************************************************************
*	sViewFulfillment
*
* Print details of one fulfillment.
***************************************************************************************************/
static FlxActBool sViewFulfillment(FlxCtuContext*	pContext,
								   FlxActLicSpc		licSpec, 
								   FlxActProdLicSpc	prodSpec,
								   void*			pData)
{
	FlxCtuViewType	viewType = *(FlxCtuViewType*)pData;

	switch (viewType)
	{
	case flxCtuViewTypeLong:
		sViewFulfillmentLong(pContext, prodSpec);
		break;
	case flxCtuViewTypeLongWithCounts:
		sViewFulfillmentCounts(pContext, prodSpec);
		break;

	case flxCtuViewTypeShort:
		/* Drop through. */
	default:
		sViewFulfillmentShort(prodSpec);
	}
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sViewFulfillmentCounts
*
* Print a fulfillment's individual counts.
***************************************************************************************************/
void sViewFulfillmentCounts(FlxCtuContext* pContext, FlxActProdLicSpc prodSpec)
{
	sViewFulfillmentLong(pContext, prodSpec);

	flxCtuPrintOutput("\n");
	sViewNameAndUint32("Concurrent", flxActCommonProdLicSpcCountGet(prodSpec, CONCURRENT));
	sViewNameAndUint32("Activatable", flxActCommonProdLicSpcCountGet(prodSpec, ACTIVATABLE));
	sViewNameAndUint32("Hybrid", flxActCommonProdLicSpcCountGet(prodSpec, HYBRID));
	sViewNameAndUint32("Concurrent overdraft", flxActCommonProdLicSpcCountGet(prodSpec, 
																			   CONCURRENT_OD));
	sViewNameAndUint32("Activatable overdraft", flxActCommonProdLicSpcCountGet(prodSpec, 
																			    ACTIVATABLE_OD));
	sViewNameAndUint32("Hybrid overdraft", flxActCommonProdLicSpcCountGet(prodSpec, HYBRID_OD));
	sViewNameAndUint32("Repair", flxActCommonProdLicSpcCountGet(prodSpec, REPAIRS));
}

/***************************************************************************************************
*	sViewVendorDictionary
*
* Print a fulfillment's vendor dictionary entries.
***************************************************************************************************/
void sViewVendorDictionary(FlxCtuContext* pContext, FlxActProdLicSpc prodSpec)
{
	uint32_t		count;
	uint32_t		index;
	const char*		pszKey;
	const char*		pszValue;

	if ( !flxActCommonProdLicSpcVendorDataGetCount(prodSpec, &count))
	{
		flxCtuContextPrintLastApiFlexError(pContext);
		return;
	}

	if (count == 0)
	{
		/* No entries. */
		return;
	}

	flxCtuPrintOutput("Vendor dictionary:\n");

	for (index = 0; index < count; ++index)
	{
		if ( !flxActCommonProdLicSpcVendorDataGetByIndex(prodSpec, index, &pszKey, &pszValue))
		{
			flxCtuContextPrintLastApiFlexError(pContext);
			return;
		}
		sViewNameAndValue(pszKey, pszValue);
	}
}

/***************************************************************************************************
*	sViewFulfillmentLong
*
* Print a fulfillment's details multi-line.
***************************************************************************************************/
static void sViewFulfillmentLong(FlxCtuContext* pContext, FlxActProdLicSpc prodSpec)
{
	flxCtuPrintOutput(
		"-------------------------------------------------------------------------------\n");
	sViewNameAndValue("Fulfillment Id", flxActCommonProdLicSpcFulfillmentIdGet(prodSpec));
	sViewNameAndValue("Fulfillment Type", sGetFulfillmentTypeString(prodSpec));
	sViewNameAndValue("Trust Flags", sGetFulfillmentTrustFlagsString(prodSpec));
	sViewNameAndValue("Status", flxActCommonProdLicSpcIsDisabled(prodSpec)? "Disabled" : "Enabled");
	sViewNameAndValue("Entitlement Id", flxActCommonProdLicSpcEntitlementIdGet(prodSpec));
	sViewNameAndValue("Product Id", flxActCommonProdLicSpcProductIdGet(prodSpec));
	sViewNameAndValue("Expiration Date", flxActCommonProdLicSpcExpDateGet(prodSpec));
	sViewNameAndValue("Server chain", flxActCommonProdLicSpcActServerChainGet(prodSpec));
	sViewNameAndFeatures("Feature Line(s)", prodSpec);
	sViewVendorDictionary(pContext, prodSpec);
}

/***************************************************************************************************
*	sViewFulfillmentShort
*
* Print a fulfillment's main details on a single line.
***************************************************************************************************/
void sViewFulfillmentShort(FlxActProdLicSpc prodSpec)
{
	FlxActBool bIsTrusted  = (flxActCommonProdLicSpcTrustFlagsGet(prodSpec) == FLAGS_FULLY_TRUSTED);
	FlxActBool bIsDisabled = flxActCommonProdLicSpcIsDisabled(prodSpec);

	FlxFulfillmentType fulfillmentType = flxActCommonProdLicSpcFulfillmentTypeGet(prodSpec);

	const char* pszFulfillmentId = flxActCommonProdLicSpcFulfillmentIdGet(prodSpec);
	const char* pszProductId	 = flxActCommonProdLicSpcProductIdGet(prodSpec);
	const char* pszExpiry		 = flxActCommonProdLicSpcExpDateGet(prodSpec);
	uint32_t nonOverdraftCount   = sGetNonOverdraftCountTotal(prodSpec);

	if (pszFulfillmentId == NULL)
		pszFulfillmentId = "<NoFid>";

	if (pszProductId == NULL)
		pszProductId = "<NoPid>";

	if (pszExpiry == NULL)
		pszExpiry = "<NoExp>";
	
	flxCtuPrintOutput("%c %c %u %-11.11s %2u %-31s %s\n",
					  bIsTrusted?  '.' : 'U',
					  bIsDisabled? 'D' : '.',
					  fulfillmentType,
					  pszExpiry,
					  nonOverdraftCount,
					  pszFulfillmentId,
					  pszProductId
				     );
}

/***************************************************************************************************
*	sViewNameAndFeatures
*
* Print a fulfillment's feature line.
***************************************************************************************************/
void sViewNameAndFeatures(const char* pszName, FlxActProdLicSpc prodSpec)
{
	const char* pszFeatureLine = flxActCommonProdLicSpcFeatureLineGet(prodSpec);

	sViewNameAndValue(pszName, "");
	if (pszFeatureLine != NULL)
	{
		flxCtuPrintOutput(pszFeatureLine);
	}
	flxCtuPrintOutput("\n");
}

/***************************************************************************************************
*	sViewNameAndValue
*
* Print name : string.
***************************************************************************************************/
void sViewNameAndValue(const char* pszName, const char* pszValue)
{
	flxCtuPrintOutput("%22.22s : %s\n", 
					  pszName,
					  (pszValue == NULL)? "" : pszValue
					 );
}

/***************************************************************************************************
*	sViewNameAndValue
*
* Print name : number.
***************************************************************************************************/
void sViewNameAndUint32(const char* pszName, uint32_t value)
{
	char pszValue[sizeof(value) * 3 + 1];
	sprintf(pszValue, "%u", value);
	sViewNameAndValue(pszName, pszValue);
}
